"""
check: functions that type check composite forms using structural subtyping
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    add_checker
    is_adjacency
    is_edge: returns whether the passed item is an edge.
    is_edges
    is_graph: returns whether the passed item is a graph.
    is_matrix
    is_node: returns whether the passed item is a node.
    is_nodes: returns whether the passed item is a collection of nodes.
    is_parallel:
    is_serial:
              
To Do:

    
"""
from __future__ import annotations
from collections.abc import (
    Collection, Hashable, MutableMapping, MutableSequence, Sequence, Set)
import inspect
import itertools
from typing import Any, Callable, Type, TYPE_CHECKING, Union

import miller

if TYPE_CHECKING:
    from . import base


def add_checker(name: str, item: Callable[[base.Graph]]) -> None:
    """Adds a checker to the local namespace.
    
    This allows the function to be found by the 'holden.classify' function and
    the 'holden.Forms.classify' class method.

    Args:
        name (str): name of the checker function. It needs to be in the 
            'is_{form}' format.
        item (Callable[[base.Graph]]): callable checker which should have a
            single parameter, item which should be a base.Graph type.
        
    """
    globals()[name] = item
    return

def is_adjacency(item: object) -> bool:
    """Returns whether 'item' is an adjacency list.

    Args:
        item (object): instance to test.

    Returns:
        bool: whether 'item' is an adjacency list.
        
    """
    if isinstance(item, MutableMapping):
        connections = list(item.values())
        nodes = list(itertools.chain.from_iterable(item.values()))
        return (
            all(isinstance(e, set) for e in connections)
            and all(is_node(item = i) for i in nodes))
    else:
        return False

def is_composite(item: Union[Type[Any], object]) -> bool:
    """Returns whether 'item' is a composite data structure.

    Args:
        item (object): instance to test.

    Returns:
        bool: whether 'item' is a composite data structure.
    
    """
    return miller.has_methods(
        item = item, 
        methods = ['add', 'delete', 'merge', 'subset'])
    
def is_edge(item: object) -> bool:
    """Returns whether 'item' is an edge.

    Args:
        item (object): instance to test.

    Returns:
        bool: whether 'item' is an edge.
        
    """
    return (
        isinstance(item, Sequence)
        and not isinstance(item, str)
        and len(item) == 2
        and is_node(item = item[0])
        and is_node(item = item[1]))
   
def is_edges(item: object) -> bool:
    """Returns whether 'item' is an edge list.

    Args:
        item (object): instance to test.

    Returns:
        bool: whether 'item' is an edge list.
    
    """
        
    return (
        isinstance(item, MutableSequence) 
        and all(is_edge(item = i) for i in item))

def is_graph(item: Union[Type[Any], object]) -> bool:
    """Returns whether 'item' is a graph.

    Args:
        item (object): instance to test.

    Returns:
        bool: whether 'item' is a graph.
    
    """
    return (
        is_composite(item = item)
        and miller.has_methods(
            item = item, 
            methods = ['connect', 'disconnect']))
    
def is_matrix(item: object) -> bool:
    """Returns whether 'item' is an adjacency matrix.

    Args:
        item (object): instance to test.

    Returns:
        bool: whether 'item' is an adjacency matrix.
        
    """
    if isinstance(item, Sequence) and len(item) == 2:
        matrix = item[0]
        labels = item[1]
        connections = list(itertools.chain.from_iterable(matrix))
        return (
            isinstance(matrix, MutableSequence)
            and isinstance(labels, MutableSequence) 
            and all(isinstance(i, MutableSequence) for i in matrix)
            and all(isinstance(n, Hashable) for n in labels)
            and all(isinstance(c, (int, float)) for c in connections))
    else:
        return False
   
def is_node(item: Union[object, Type[Any]]) -> bool:
    """Returns whether 'item' is a node.

    Args:
        item (Union[object, Type[Any]]): instance or class to test.

    Returns:
        bool: whether 'item' is a node.
    
    """
    if inspect.isclass(item):
        return issubclass(item, Hashable)
    else:
        return isinstance(item, Hashable)

def is_nodes(item: object) -> bool:
    """Returns whether 'item' is a collection of nodes.

    Args:
        item (object): instance to test.

    Returns:
        bool: whether 'item' is a collection of nodes.
    
    """
    return (
        isinstance(item, Collection) and all(is_node(item = i) for i in item))

def is_parallel(item: object) -> bool:
    """Returns whether 'item' is sequence of parallel.

    Args:
        item (object): instance to test.

    Returns:
        bool: whether 'item' is a sequence of parallel.
    
    """
    return (
        isinstance(item, MutableSequence)
        and all(is_serial(item = i) for i in item))

def is_serial(item: object) -> bool:
    """Returns whether 'item' is a serial.

    Args:
        item (object): instance to test.

    Returns:
        bool: whether 'item' is a serial.
    
    """
    return (
        isinstance(item, MutableSequence)
        and all(is_node(item = i) for i in item)
        and all(not isinstance(i, tuple) for i in item))


"""
composites: base types of other composite data structures
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    Parallel (camina.Listing, traits.Directed, base.Composite):
    Serial (camina.Hybrid, traits.Directed, base.Composite):
                  
To Do:
    Complete Tree class and related functions
    Integrate Kinds system when it is finished
  
"""
from __future__ import annotations
from collections.abc import Hashable, MutableSequence, Sequence
import copy
import dataclasses
from typing import Any, Optional, Union

import camina

from . import base
from . import check
from . import report
from . import traits
from . import traverse

    
@dataclasses.dataclass
class Parallel(camina.Listing, traits.Directed, base.Composite):
    """Base class for a list of serial composites.
    
    Args:
        contents (MutableSequence[Serial]): Listing of Serial instances. 
            Defaults to an empty list.
                                      
    """   
    contents: MutableSequence[Serial] = dataclasses.field(
        default_factory = list)
                                
    """ Properties """
    
    @property
    def endpoint(self) -> MutableSequence[Hashable]:
        """Returns the endpoints of the stored composite."""
        return report.get_endpoints_parallel(item = self)
                    
    @property
    def root(self) -> MutableSequence[Hashable]:
        """Returns the roots of the stored composite."""
        return report.get_roots_parallel(item = self)

    """ Public Methods """
    
    def walk(
        self, 
        start: Optional[Hashable] = None, 
        stop: Optional[Hashable] = None) -> Parallel:
        """Returns all paths in graph from 'start' to 'stop'.
        
        Args:
            start (Hashable): node to start paths from.
            stop (Hashable): node to stop paths.
            
        Returns:
            Paralle: a list of possible paths (each path is a list nodes) from 
                'start' to 'stop'.
            
        """
        if start is None:
            root = self.root
        else:
            root = camina.listify(start)
        if stop is None:
            endpoint = self.endpoint
        else:
            endpoint = self.camina.listify(stop)
        return traverse.walk_parallel(
            item = self, 
            start = root, 
            stop = endpoint) 
    
    """ Private Methods """   
    
    def _add(self, item: Hashable, *args: Any, **kwargs: Any) -> None:
        """Adds node to the stored composite.
                   
        Args:
            item (Hashable): node to add to the stored composite.
            
        """
        self.contents.append(item)
        return
      
    def _delete(self, item: Hashable, *args: Any, **kwargs: Any) -> None:
        """Deletes node from the stored composite.
                
        Args:
            item (Hashable): node to delete from 'contents'.
        
            
        """
        del self.contents[item]
        return

    def _merge(self, item: base.Composite, *args: Any, **kwargs: Any) -> None:
        """Combines 'item' with the stored composite.

        Subclasses must provide their own specific methods for merging with
        another composite. The provided 'merge' method offers all of the error 
        checking. Subclasses just need to provide the mechanism for merging 
        ithout worrying about validation or error-checking.
        
        Args:
            item (base.Composite): another Composite object to add to the 
                stored composite.
                
        """
        other = base.transform(
            item = item, 
            output = 'parallel', 
            raise_same_error = False)
        for serial in other:
            self.contents.append(serial)
        return
    
    def _subset(
        self, 
        include: Union[Hashable, Sequence[Hashable]] = None,
        exclude: Union[Hashable, Sequence[Hashable]] = None) -> Parallel:
        """Returns a new composite without a subset of 'contents'.

        Subclasses must provide their own specific methods for deleting a single
        edge. Subclasses just need to provide the mechanism for returning a
        subset without worrying about validation or error-checking.
        
        Args:
            include (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should be included in the new composite.
            exclude (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should not be included in the new composite.

        Returns:
           Parallel: with only selected nodes and edges.
            
        """
        raise NotImplementedError   
                               
    """ Dunder Methods """
        
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        """Returns whether 'instance' meets criteria to be a subclass.

        Args:
            instance (object): item to test as an instance.

        Returns:
            bool: whether 'instance' meets criteria to be a subclass.
            
        """
        return check.is_parallel(item = instance)
     
    
@dataclasses.dataclass
class Serial(camina.Hybrid, traits.Directed, base.Composite):
    """Base class for serial composites.
    
    Args:
        contents (MutableSequence[Hashable]): list of nodes. Defaults to 
            an empty list.
                                      
    """   
    contents: MutableSequence[Hashable] = dataclasses.field(
        default_factory = list)
                   
    """ Properties """

    @property
    def endpoint(self) -> MutableSequence[Hashable]:
        """Returns the endpoints of the stored composite."""
        return report.get_endpoints_serial(item = self)
                    
    @property
    def root(self) -> MutableSequence[Hashable]:
        """Returns the roots of the stored composite."""
        return report.get_roots_serial(item = self)

    """ Public Methods """
    
    def walk(
        self, 
        start: Optional[Hashable] = None, 
        stop: Optional[Hashable] = None) -> Parallel:
        """Returns all paths in graph from 'start' to 'stop'.
        
        Args:
            start (Hashable): node to start paths from.
            stop (Hashable): node to stop paths.
            
        Returns:
            Paralle: a list of possible paths (each path is a list nodes) from 
                'start' to 'stop'.
            
        """
        if start is None:
            start = self.root[0]
        if stop is None:
            stop = self.endpoint[0]
        return traverse.walk_serial(item = self, start = start, stop = stop) 
        
    """ Private Methods """   
    
    def _add(self, item: Hashable, *args: Any, **kwargs: Any) -> None:
        """Adds node to the stored composite.
                   
        Args:
            item (Hashable): node to add to the stored composite.
            
        """
        self.contents.append(item)
        return
      
    def _delete(self, item: Hashable, *args: Any, **kwargs: Any) -> None:
        """Deletes node from the stored composite.
                
        Args:
            item (Hashable): node to delete from 'contents'.
        
            
        """
        del self.contents[item]
        return

    def _merge(self, item: base.Composite, *args: Any, **kwargs: Any) -> None:
        """Combines 'item' with the stored composite.

        Subclasses must provide their own specific methods for merging with
        another composite. The provided 'merge' method offers all of the error 
        checking. Subclasses just need to provide the mechanism for merging 
        ithout worrying about validation or error-checking.
        
        Args:
            item (base.Composite): another Composite object to add to the 
                stored composite.
                
        """
        other = base.transform(
            item = item, 
            output = 'serial', 
            raise_same_error = False)
        self.contents.extend(other)
        return
    
    def _subset(
        self, 
        include: Union[Hashable, Sequence[Hashable]] = None,
        exclude: Union[Hashable, Sequence[Hashable]] = None) -> Serial:
        """Returns a new composite without a subset of 'contents'.

        Subclasses must provide their own specific methods for deleting a single
        edge. Subclasses just need to provide the mechanism for returning a
        subset without worrying about validation or error-checking.
        
        Args:
            include (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should be included in the new composite.
            exclude (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should not be included in the new composite.

        Returns:
           Serial: with only selected nodes and edges.
            
        """
        if include:
            new_serial = [i for i in self.contents if i in include]  
        else:
            new_serial = copy.deepcopy(self.contents)
        if exclude:
            new_serial = [i for i in self.contents if i not in exclude]
        return self.__class__(contents = new_serial)
                        
    """ Dunder Methods """
        
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        """Returns whether 'instance' meets criteria to be a subclass.

        Args:
            instance (object): item to test as an instance.

        Returns:
            bool: whether 'instance' meets criteria to be a subclass.
            
        """
        return check.is_serial(item = instance)      


""" Type Checkers """

    
# @dataclasses.dataclass # type: ignore
# class Tree(camina.Hybrid, traits.Directed, base.Composite):
#     """Base class for an tree data structures.
    
#     The Tree class uses a Hybrid instead of a linked list for storing children
#     nodes to allow easier access of nodes further away from the root. For
#     example, a user might use 'a_tree["big_branch"]["small_branch"]["a_leaf"]' 
#     to access a desired node instead of 'a_tree[2][0][3]' (although the latter
#     access technique is also supported).

#     Args:
#         contents (MutableSequence[Node]): list of stored Tree or other 
#             Node instances. Defaults to an empty list.
#         name (Optional[str]): name of Tree node. Defaults to None.
#         parent (Optional[Tree]): parent Tree, if any. Defaults to None.
#         default_factory (Optional[Any]): default value to return or default 
#             function to call when the 'get' method is used. Defaults to None. 
              
#     """
#     contents: MutableSequence[Hashable] = dataclasses.field(
#         default_factory = list)
#     name: Optional[str] = None
#     parent: Optional[Tree] = None
#     default_factory: Optional[Any] = None
                    
#     """ Properties """
        
#     @property
#     def children(self) -> MutableSequence[Hashable]:
#         """Returns child nodes of this Node."""
#         return self.contents
    
#     @children.setter
#     def children(self, value: MutableSequence[Hashable]) -> None:
#         """Sets child nodes of this Node."""
#         if camina.is_sequence(value):
#             self.contents = value
#         else:
#             self.contents = [value]
#         return

#     @property
#     def endpoint(self) -> Union[Hashable, Collection[Hashable]]:
#         """Returns the endpoint(s) of the stored composite."""
#         if not self.contents:
#             return self
#         else:
#             return self.contents[0].endpoint
 
#     @property
#     def root(self) -> Union[Hashable, Collection[Hashable]]:
#         """Returns the root(s) of the stored composite."""
#         if self.parent is None:
#             return self
#         else:
#             return self.parent.root  
                                
#     """ Dunder Methods """
        
#     @classmethod
#     def __instancecheck__(cls, instance: object) -> bool:
#         """Returns whether 'instance' meets criteria to be a subclass.

#         Args:
#             instance (object): item to test as an instance.

#         Returns:
#             bool: whether 'instance' meets criteria to be a subclass.
            
#         """
#         return is_tree(item = instance)

#     def __missing__(self) -> Tree:
#         """Returns an empty tree if one does not exist.

#         Returns:
#             Tree: an empty instance of Tree.
            
#         """
#         return self.__class__()


# def is_tree(item: object) -> bool:
#     """Returns whether 'item' is a tree.

#     Args:
#         item (object): instance to test.

#     Returns:
#         bool: whether 'item' is a tree.
    
#     """
#     return (
#         isinstance(item, MutableSequence)
#         and all(isinstance(i, (MutableSequence, Hashable)) for i in item)) 
    
# def is_forest(item: object) -> bool:
#     """Returns whether 'item' is a dict of tree.

#     Args:
#         item (object): instance to test.

#     Returns:
#         bool: whether 'item' is a dict of tree.
    
#     """
#     return (
#         isinstance(item, MutableMapping)
#         and all(base.is_node(item = i) for i in item.keys())
#         and all(is_tree(item = i) for i in item.values())) 


# # @functools.singledispatch 
# def to_tree(item: Any) -> graphs.Tree:
#     """Converts 'item' to a Tree.
    
#     Args:
#         item (Any): item to convert to a Tree.

#     Raises:
#         TypeError: if 'item' is a type that is not registered.

#     Returns:
#         form.Tree: derived from 'item'.

#     """
#     if check.is_tree(item = item):
#         return item
#     else:
#         raise TypeError(
#             f'item cannot be converted because it is an unsupported type: '
#             f'{type(item).__name__}')

# # @to_tree.register # type: ignore 
# def matrix_to_tree(item: graphs.Matrix) -> graphs.Tree:
#     """Converts 'item' to a Tree.
    
#     Args:
#         item (form.Matrix): item to convert to a Tree.

#     Raises:
#         TypeError: if 'item' is a type that is not registered.

#     Returns:
#         form.Tree: derived from 'item'.

#     """
#     tree = {}
#     for node in item:
#         children = item[:]
#         children.remove(node)
#         tree[node] = matrix_to_tree(children)
#     return tree
        


"""
report: inspection tools for composite data structures.
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    get_endpoints_adjacency
    get_roots_adjacency
    get_endpoints_edges
    get_roots_edges
    get_endpoints_matrix
    get_roots_matrix  
           
To Do:
    Implement remaining functions.
    
"""
from __future__ import annotations
from collections.abc import Hashable, MutableSequence
# import functools
import itertools
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from . import composites
    from . import graphs

    
""" Introspection Tools """

def get_endpoints_adjacency(
    item: graphs.Adjacency) -> MutableSequence[Hashable]:
    """Returns the endpoints in 'item'.

    Args:
        item (graphs.Adjacency): adjacency list object to examine.

    Returns:
        MutableSequence[Hashable]: list of endpoints.
        
    """  
    return [k for k in item.keys() if not item[k]]

def get_roots_adjacency(item: graphs.Adjacency) -> MutableSequence[Hashable]:
    """Returns the roots in 'item'.

    Args:
        item (graphs.Adjacency): adjacency list object to examine.

    Returns:
        MutableSequence[Hashable]: list of roots.
        
    """  
    stops = list(itertools.chain.from_iterable(item.values()))
    return [k for k in item.keys() if k not in stops]  

def get_endpoints_edges(item: graphs.Edges) -> MutableSequence[Hashable]:
    """Returns the endpoints in 'item'.

    Args:
        item (graphs.Edges): edge list object to examine.

    Returns:
        MutableSequence[Hashable]: list of endpoints.
        
    """  
    raise NotImplementedError

def get_roots_edges(item: graphs.Edges) -> MutableSequence[Hashable]:
    """Returns the roots in 'item'.

    Args:
        item (graphs.Edges): edge list object to examine.

    Returns:
        MutableSequence[Hashable]: list of roots.
        
    """  
    raise NotImplementedError    

def get_endpoints_matrix(item: graphs.Matrix) -> MutableSequence[Hashable]:
    """Returns the endpoints in 'item'.

    Args:
        item (graphs.Matrix): adjacency matrix object to examine.

    Returns:
        MutableSequence[Hashable]: list of endpoints.
        
    """  
    raise NotImplementedError

def get_roots_matrix(item: graphs.Matrix) -> MutableSequence[Hashable]:
    """Returns the roots in 'item'.

    Args:
        item (graphs.Matrix): adjacency matrix object to examine.

    Returns:
        MutableSequence[Hashable]: list of roots.
        
    """  
    raise NotImplementedError    

def get_endpoints_parallel(
    item: composites.Parallel) -> MutableSequence[Hashable]:
    """Returns the endpoints in 'item'.

    Args:
        item (composites.Parallel): parallel object to examine.

    Returns:
        MutableSequence[Hashable]: list of endpoints.
        
    """  
    return [p[-1] for p in item]

def get_roots_parallel(item: composites.Parallel) -> MutableSequence[Hashable]:
    """Returns the roots in 'item'.

    Args:
        item (composites.Parallel): parallel object to examine.

    Returns:
        MutableSequence[Hashable]: list of roots.
        
    """  
    return [p[0] for p in item]   

def get_endpoints_serial(item: composites.Serial) -> MutableSequence[Hashable]:
    """Returns the endpoints in 'item'.

    Args:
        item (composites.Serial): serial object to examine.

    Returns:
        MutableSequence[Hashable]: list of endpoints.
        
    """  
    return [item[-1]]

def get_roots_serial(item: composites.Serial) -> MutableSequence[Hashable]:
    """Returns the roots in 'item'.

    Args:
        item (composites.Serial): serial object to examine.

    Returns:
        MutableSequence[Hashable]: list of roots.
        
    """  
    return [item[0]]



"""
traverse: internal storage formats for graphs
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    walk_adjacency
    walk_edges
    walk_matrix
    walk_parallel
    walk_serial
          
To Do:
    Complete not implemented functions
    For adjacency matrix walk, consider the efficient approach here:
        https://www.geeksforgeeks.org/count-possible-paths-source-destination-exactly-k-edges/
    
"""
from __future__ import annotations
from collections.abc import Hashable, Sequence
from typing import Any, ClassVar, Optional, Type, TYPE_CHECKING, Union

if TYPE_CHECKING:
    from . import composites
    from . import graphs


def walk_adjacency(
    item: graphs.Adjacency, 
    start: Hashable, 
    stop: Hashable,
    path: Optional[Sequence[Hashable]] = None) -> Sequence[Hashable]:
    """Returns all paths in 'item' from 'start' to 'stop'.

    The code here is adapted from: https://www.python.org/doc/essays/graphs/
    
    Args:
        item (graphs.Adjacency): item in which to find paths.
        start (Hashable): node to start paths from.
        stop (Hashable): node to stop paths.
        path (Optional[Sequence[Hashable]]): a path from 'start' to 'stop'. This
            is used for recursion within the function to accumulate all possible
            paths. Defaults to None. 

    Returns:
        Sequence[Hashable]: a list of possible paths (each path is a list nodes) 
            from 'start' to 'stop'.
        
    """            
    if path is None:
        path = []
    path = path + [start]
    if start == stop:
        return [path]
    if start not in item:
        return []
    paths = []
    for node in item[start]:
        if node not in path:
            new_paths = walk_adjacency(
                item = item,
                start = node, 
                stop = stop, 
                path = path)
            for new_path in new_paths:
                paths.append(new_path)
    return paths

def walk_edges(
    item: graphs.Edges, 
    start: Hashable, 
    stop: Hashable,
    path: Optional[Sequence[Hashable]] = None) -> Sequence[Hashable]:
    """Returns all paths in 'item' from 'start' to 'stop'.

    Args:
        item (graphs.Edges): item in which to find paths.
        start (Hashable): node to start paths from.
        stop (Hashable): node to stop paths.
        path (Optional[Sequence[Hashable]]): a path from 'start' to 'stop'. 
            Defaults to None. 

    Returns:
        Sequence[Hashable]: a list of possible paths (each path is a list nodes) 
            from 'start' to 'stop'.
        
    """            
    raise NotImplementedError  

def walk_matrix(
    item: graphs.Matrix, 
    start: Hashable, 
    stop: Hashable,
    path: Optional[Sequence[Hashable]] = None) -> Sequence[Hashable]:
    """Returns all paths in 'item' from 'start' to 'stop'.

    Args:
        item (graphs.Matrix): item in which to find paths.
        start (Hashable): node to start paths from.
        stop (Hashable): node to stop paths.
        path (Optional[Sequence[Hashable]]): a path from 'start' to 'stop'. 
            Defaults to None. 

    Returns:
        Sequence[Hashable]: a list of possible paths (each path is a list nodes) 
            from 'start' to 'stop'.
        
    """            
    raise NotImplementedError         

def walk_parallel(
    item: composites.Parallel, 
    start: Hashable, 
    stop: Hashable) -> Sequence[Hashable]:
    """Returns all paths in 'item' from 'start' to 'stop'.
    
    Args:
        item (composites.Parallel): item in which to find paths.
        start (Hashable): node to start paths from.
        stop (Hashable): node to stop paths.

    Returns:
        Sequence[Hashable]: a list of possible paths (each path is a list nodes) 
            from 'start' to 'stop'.
        
    """            
    return [walk_serial(item = p, start = start, stop = stop) for p in item]

def walk_serial(
    item: composites.Serial, 
    start: Hashable, 
    stop: Hashable) -> Sequence[Hashable]:
    """Returns all paths in 'item' from 'start' to 'stop'.
    
    Args:
        item (composites.Serial): item in which to find paths.
        start (Hashable): node to start paths from.
        stop (Hashable): node to stop paths.

    Returns:
        Sequence[Hashable]: a list of possible paths (each path is a list nodes) 
            from 'start' to 'stop'.
        
    """            
    index_start = item.index(start)
    index_stop = item.index(stop)
    if index_stop > len(item):
        path = item[index_start:]
    else:
        path = item[index_start:index_stop]
    return path


"""
export: functions to export composite data structures to other formats
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:

           
To Do:

    
"""
from __future__ import annotations
import pathlib
from typing import Any, Callable, Optional

from . import base
from . import traits


_LINE_BREAK = '\n'
_DIRECTED_LINK = '->'
_UNDIRECTED_LINK = '--'

def to_dot(
    item: base.Composite, 
    path: Optional[str | pathlib.Path] = None,
    name: Optional[str] = None,
    settings: Optional[dict[str, Any]] = None) -> str:
    """Converts 'item' to a dot format.

    Args:
        item (base.Composite): item to convert to a dot format.
        path (Optional[str | pathlib.Path]): path to export 'item' to. Defaults
            to None.
        name (Optional[str]): name of 'item' to put in the dot str. Defaults to
            None.
        settings (Optional[dict[str, Any]]): any global settings to add to the
            dot graph. Defaults to None.

    Returns:
        str: composite object in graphviz dot format.

    """
    edges = base.transform(
        item = item, 
        output = 'edges', 
        raise_same_error = False)
    name = name or 'holden'
    if isinstance(item, traits.Directed):
        dot = 'digraph '
        link = _DIRECTED_LINK
    else:
        dot = 'graph '
        link = _UNDIRECTED_LINK
    dot = dot +  name + ' {\n'
    if settings is not None:
        for key, value in settings.items():
            dot = dot + f'{key}={value};{_LINE_BREAK}'
    for edge in edges:
        dot = dot + f'{edge[0]} {link} {edge[1]}{_LINE_BREAK}'
    dot = dot + '}'
    if path is not None:
        with open(path, 'w') as a_file:
            a_file.write(dot)
        a_file.close()
    return dot
 


"""
defaults: settings for default classes for each graph form
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

This module does not presently have any function. It is intended to be used by
other packages which wish to use holden tools with other base type classes.

Contents:
    
          
To Do:

    
"""
from __future__ import annotations
from typing import Type

from . import base
from . import composites
from . import graphs


_BASE_ADJACENCY: Type[base.Graph] = graphs.Adjacency
_BASE_EDGES: Type[base.Graph] = graphs.Edges
_BASE_MATRIX: Type[base.Graph] = graphs.Matrix
_BASE_PARALLEL: Type[base.Graph] = composites.Parallel
_BASE_SERIAL: Type[base.Graph] = composites.Serial

def set_base(name: str, value: Type[base.Graph]) -> None:
    """Sets default base class for a form of graph.

    Args:
        name (str): name of form to set.
        value (Type[base.Graph]): Graph subclass to use as the base type for
            the 'name' form.
            
    """
    variable = f'(_BASE_{name.upper()})'
    globals()[variable] = value
    return

                     


"""
workshop: functions to change the internal storage format for a data structure
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    add_transformer
    adjacency_to_edges
    adjacency_to_matrix
    adjacency_to_serial
    adjacency_to_parallel
    edges_to_adjacency
    edges_to_matrix
    edges_to_serial
    edges_to_parallel
    matrix_to_adjacency
    matrix_to_edges
    matrix_to_serial
    matrix_to_parallel
    serial_to_adjacency
    serial_to_edges
    serial_to_matrix
    serial_to_parallel
    parallel_to_adjacency
    parallel_to_edges
    parallel_to_matrix
    parallel_to_serial
           
To Do:
    Implement remaining functions.
    
"""
from __future__ import annotations
import collections
from collections.abc import (
    Collection, Hashable, MutableMapping, MutableSequence, Sequence, Set)
# import functools
import itertools
from typing import Callable, TYPE_CHECKING

import camina

from . import check
from . import traverse

if TYPE_CHECKING:
    from . import base
    from . import composites
    from . import graphs


""" Transformers """

def add_transformer(name: str, item: Callable[[base.Graph]]) -> None:
    """Adds a transformer to the local namespace.
    
    This allows the function to be found by the 'transform' function.

    Args:
        name (str): name of the transformer function. It needs to be in the 
            '{input form}_to_{output form}' format.
        item (Callable[[base.Graph]]): callable transformer which should have a
            single parameter, item which should be a base.Graph type.
        
    """
    globals()[name] = item
    return

# @to_edges.register # type: ignore
def adjacency_to_edges(item: graphs.Adjacency) -> graphs.Edges:
    """Converts 'item' to an graphs.Edges.
    
    Args:
        item (graphs.Adjacency): item to convert to an graphs.Edges.

    Returns:
        graphs.Edges: derived from 'item'.

    """ 
    edges = []
    for node, connections in item.items():
        for connection in connections:
            edges.append(tuple([node, connection]))
    return tuple(edges)

# @to_matrix.register # type: ignore 
def adjacency_to_matrix(item: graphs.Adjacency) -> graphs.Matrix:
    """Converts 'item' to a graphs.Matrix.
    
    Args:
        item (graphs.Adjacency): item to convert to a graphs.Matrix.

    Returns:
        graphs.Matrix: derived from 'item'.

    """ 
    names = list(item.keys())
    matrix = []
    for i in range(len(item)): 
        matrix.append([0] * len(item))
        for j in item[i]:
            matrix[i][j] = 1
    return tuple([matrix, names])    

# @to_parallel.register # type: ignore 
def adjacency_to_parallel(item: graphs.Adjacency) -> composites.Parallel:
    """Converts 'item' to a Serial.
    
    Args:
        item (graphs.Adjacency): item to convert to a Serial.

    Returns:
        Serial: derived from 'item'.

    """
    roots = get_roots_adjacency(item = item)
    endpoints = get_endpoints_adjacency(item = item)
    all_paths = []
    for start in roots:
        for end in endpoints:
            paths = traverse.walk_adjacency(
                item = item, 
                start = start, 
                stop = end)
            if paths:
                if all(isinstance(path, Hashable) for path in paths):
                    all_paths.append(paths)
                else:
                    all_paths.extend(paths)
    return all_paths

# @to_serial.register # type: ignore 
def adjacency_to_serial(item: graphs.Adjacency) -> composites.Serial:
    """Converts 'item' to a Serial.
    
    Args:
        item (graphs.Adjacency): item to convert to a Serial.

    Returns:
        Serial: derived from 'item'.

    """ 
    all_parallel = adjacency_to_parallel(item = item)
    if len(all_parallel) == 1:
        return all_parallel[0]
    else:
        return list(itertools.chain.from_iterable(all_parallel))
       
# @to_adjacency.register # type: ignore
def edges_to_adjacency(item: graphs.Edges) -> graphs.Adjacency:
    """Converts 'item' to an graphs.Adjacency.

    Args:
        item (graphs.Edges): item to convert to an graphs.Adjacency.

    Returns:
        graphs.Adjacency: derived from 'item'.

    """
    adjacency = collections.defaultdict(set)
    for edge_pair in item:
        if edge_pair[0] not in adjacency:
            adjacency[edge_pair[0]] = {edge_pair[1]}
        else:
            adjacency[edge_pair[0]].add(edge_pair[1])
        if edge_pair[1] not in adjacency:
            adjacency[edge_pair[1]] = set()
    return adjacency

# @to_matrix.register # type: ignore 
def edges_to_matrix(item: graphs.Edges) -> graphs.Matrix:
    """Converts 'item' to a graphs.Matrix.

    Args:
        item (graphs.Edges): item to convert to a graphs.Matrix.

    Returns:
        graphs.Matrix: derived from 'item'.

    """
    raise NotImplementedError
 
# @to_parallel.register # type: ignore 
def edges_to_parallel(item: graphs.Edges) -> composites.Parallel:
    """Converts 'item' to a Parallel.

    Args:
        item (graphs.Edges): item to convert to a Parallel.

    Returns:
        Parallel: derived from 'item'.

    """
    raise NotImplementedError
   
# @to_serial.register # type: ignore 
def edges_to_serial(item: graphs.Edges) -> composites.Serial:
    """Converts 'item' to a Serial.

    Args:
        item (graphs.Edges): item to convert to a Serial.

    Returns:
        Serial: derived from 'item'.

    """
    raise NotImplementedError

# @to_adjacency.register # type: ignore 
def matrix_to_adjacency(item: graphs.Matrix) -> graphs.Adjacency:
    """Converts 'item' to an graphs.Adjacency.

    Args:
        item (graphs.Matrix): item to convert to an graphs.Adjacency.

    Returns:
        graphs.Adjacency: derived from 'item'.

    """  
    matrix = item[0]
    names = item[1]
    name_mapping = dict(zip(range(len(matrix)), names))
    raw_adjacency = {
        i: [j for j, adjacent in enumerate(row) if adjacent] 
        for i, row in enumerate(matrix)}
    adjacency = collections.defaultdict(set)
    for key, value in raw_adjacency.items():
        new_key = name_mapping[key]
        new_values = set()
        for edge in value:
            new_values.add(name_mapping[edge])
        adjacency[new_key] = new_values
    return adjacency
    
# @to_edges.register # type: ignore 
def matrix_to_edges(item: graphs.Matrix) -> graphs.Edges:
    """Converts 'item' to an graphs.Edges.

    Args:
        item (graphs.Matrix): item to convert to an graphs.Edges.

    Returns:
        graphs.Edges: derived from 'item'.

    """
    matrix = item[0]
    labels = item[1]
    edges = []
    for i in enumerate(matrix):
        for j in enumerate(matrix):
            if matrix[i][j] > 0:
                edges.append((labels[i], labels[j]))
    return edges
 
# @to_parallel.register # type: ignore 
def matrix_to_parallel(item: graphs.Matrix) -> composites.Parallel:
    """Converts 'item' to a Parallel.

    Args:
        item (graphs.Matrix): item to convert to a Parallel.

    Returns:
        Parallel: derived from 'item'.

    """
    raise NotImplementedError
   
# @to_serial.register # type: ignore 
def matrix_to_serial(item: graphs.Matrix) -> composites.Serial:
    """Converts 'item' to a Serial.

    Args:
        item (graphs.Matrix): item to convert to a Serial.

    Returns:
        Serial: derived from 'item'.

    """
    raise NotImplementedError

# @to_adjacency.register # type: ignore 
def parallel_to_adjacency(item: composites.Parallel) -> graphs.Adjacency:
    """Converts 'item' to an graphs.Adjacency.

    Args:
        item (Parallel): item to convert to an graphs.Adjacency.

    Returns:
        graphs.Adjacency: derived from 'item'.

    """
    adjacency = collections.defaultdict(set)
    for serial in item:
        pipe_adjacency = serial_to_adjacency(item = serial)
        for key, value in pipe_adjacency.items():
            if key in adjacency:
                for inner_value in value:
                    if inner_value not in adjacency:
                        adjacency[key].add(inner_value)
            else:
                adjacency[key] = value
    return adjacency  
    
# @to_edges.register # type: ignore 
def parallel_to_edges(item: composites.Parallel) -> graphs.Edges:
    """Converts 'item' to an graphs.Edges.

    Args:
        item (Parallel): item to convert to an graphs.Edges.

    Returns:
        graphs.Edges: derived from 'item'.

    """
    raise NotImplementedError
 
# @to_matrix.register # type: ignore 
def parallel_to_matrix(item: composites.Parallel) -> graphs.Matrix:
    """Converts 'item' to a graphs.Matrix.

    Args:
        item (Parallel): item to convert to a graphs.Matrix.

    Returns:
        graphs.Matrix: derived from 'item'.

    """
    raise NotImplementedError
   
# @to_serial.register # type: ignore 
def parallel_to_serial(item: composites.Parallel) -> composites.Serial:
    """Converts 'item' to a Serial.

    Args:
        item (Parallel): item to convert to a Serial.

    Returns:
        Serial: derived from 'item'.

    """
    raise NotImplementedError
    
# @to_adjacency.register # type: ignore 
def serial_to_adjacency(item: composites.Serial) -> graphs.Adjacency:
    """Converts 'item' to an graphs.Adjacency.

    Args:
        item (Serial): item to convert to an graphs.Adjacency.

    Returns:
        graphs.Adjacency: derived from 'item'.

    """
    if check.is_parallel(item = item):
        return parallel_to_adjacency(item = item)
    else:
        if not isinstance(item, (Collection)) or isinstance(item, str):
            item = [item]
        adjacency = collections.defaultdict(set)
        if len(item) == 1:
            adjacency.update({item[0]: set()})
        else:
            edges = list(camina.windowify(item, 2))
            for edge_pair in edges:
                if edge_pair[0] in adjacency:
                    adjacency[edge_pair[0]].add(edge_pair[1])
                else:
                    adjacency[edge_pair[0]] = {edge_pair[1]} 
        return adjacency
    
# @to_edges.register # type: ignore 
def serial_to_edges(item: composites.Serial) -> graphs.Edges:
    """Converts 'item' to an graphs.Edges.

    Args:
        item (Serial): item to convert to an graphs.Edges.

    Returns:
        graphs.Edges: derived from 'item'.

    """
    raise NotImplementedError
    
# @to_matrix.register # type: ignore 
def serial_to_matrix(item: composites.Serial) -> graphs.Matrix:
    """Converts 'item' to a graphs.Matrix.

    Args:
        item (Serial): item to convert to a graphs.Matrix.

    Returns:
        graphs.Matrix: derived from 'item'.

    """
    raise NotImplementedError
 
# @to_parallel.register # type: ignore 
def serial_to_parallel(item: composites.Serial) -> composites.Parallel:
    """Converts 'item' to a Parallel.

    Args:
        item (Serial): item to convert to a Parallel.

    Returns:
        Parallel: derived from 'item'.

    """
    raise NotImplementedError

    
""" Introspection Tools """

def get_endpoints_adjacency(item: graphs.Adjacency) -> MutableSequence[Hashable]:
    """Returns the endpoints in 'item'."""
    return [k for k in item.keys() if not item[k]]

def get_roots_adjacency(item: graphs.Adjacency) -> MutableSequence[Hashable]:
    """Returns the roots in 'item'."""
    stops = list(itertools.chain.from_iterable(item.values()))
    return [k for k in item.keys() if k not in stops]  

""" 
These are functions design to implement a dispatch system for the form
tranformers. However, functools.singledispatch has some shortcomings. If a new
dispatch system is developed in camina or the functools decorator is improved,
these functions may be restored to allow more flexible function calls.

"""          
# @functools.singledispatch
# def to_adjacency(item: object) -> graphs.Adjacency:
#     """Converts 'item' to an graphs.Adjacency.
    
#     Args:
#         item (object): item to convert to an graphs.Adjacency.

#     Raises:
#         TypeError: if 'item' is a type that is not registered with the 
#         dispatcher.

#     Returns:
#         graphs.Adjacency: derived from 'item'.

#     """
#     if is_adjacency(item = item):
#         return item
#     else:
#         raise TypeError(
#             f'item cannot be converted because it is an unsupported type: '
#             f'{type(item).__name__}')

# @functools.singledispatch  
# def to_edges(item: object) -> graphs.Edges:
#     """Converts 'item' to an graphs.Edges.
    
#     Args:
#         item (object): item to convert to an graphs.Edges.

#     Raises:
#         TypeError: if 'item' is a type that is not registered.

#     Returns:
#         graphs.Edges: derived from 'item'.

#     """
#     if is_edges(item = item):
#         return item
#     else:
#         raise TypeError(
#             f'item cannot be converted because it is an unsupported type: '
#             f'{type(item).__name__}')

# @functools.singledispatch   
# def to_matrix(item: object) -> graphs.Matrix:
#     """Converts 'item' to a graphs.Matrix.
    
#     Args:
#         item (object): item to convert to a graphs.Matrix.

#     Raises:
#         TypeError: if 'item' is a type that is not registered.

#     Returns:
#         graphs.Matrix: derived from 'item'.

#     """
#     if is_matrix(item = item):
#         return item
#     else:
#         raise TypeError(
#             f'item cannot be converted because it is an unsupported type: '
#             f'{type(item).__name__}')
                


"""
graphs: graphs with different internal storage formats
Corey Rayburn Yung <coreyrayburnyung@gmail.com>
Copyright 2020-2022, Corey Rayburn Yung
License: Apache-2.0

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.

Contents:
    Adjacency (base.Graph, camina.Dictionary): a graph stored as an adjacency 
        list.
    Edges (base.Graph, camina.Listing): a graph stored as an edge list.
    Matrix (base.Graph, camina.Listing): a graph stored as an adjacency matrix.
         
To Do:
    Add the remainder of the conversion methods between different forms
    Add private methods that currently raise NotImplementedError
    Integrate Kinds system when it is finished
    
"""
from __future__ import annotations
import collections
from collections.abc import (
    Hashable, MutableMapping, MutableSequence, Sequence, Set)
import copy
import dataclasses
from typing import Any, ClassVar, Optional, Type, TYPE_CHECKING, Union

import camina

from . import base
from . import check


""" Graph Form Base Classes """

@dataclasses.dataclass
class Adjacency(base.Graph, camina.Dictionary):
    """Base class for adjacency-list graphs.
    
    Args:
        contents (MutableMapping[Hashable, set[Hashable]]): keys are hashable 
            representations of nodes. Values are the nodes to which the key node
            are connected. In a directed graph, the key node is assumed to come
            before the value node in order. Defaults to a defaultdict that has a 
            set for its value type.
                                      
    """  
    contents: MutableMapping[Hashable, set[Hashable]] = dataclasses.field(
            default_factory = lambda: collections.defaultdict(set))

    """ Public Methods """   
    
    def _add(self, item: Hashable, *args: Any, **kwargs: Any) -> None:
        """Adds node to the stored graph.
                   
        Args:
            item (Hashable): node to add to the stored graph.
            
        """
        self.contents[item] = set()
        return
        
    def _connect(self, item: base.Edge, *args: Any, **kwargs: Any) -> None:
        """Adds edge to the stored graph.
        
        Args:
            item (Edge): edge to add to the stored graph.
            
        """
        self.contents[item[0]].add(item[1])
        return
      
    def _delete(self, item: Hashable, *args: Any, **kwargs: Any) -> None:
        """Deletes node from the stored graph.
          
        Args:
            item (Hashable): node to delete from 'contents'.
        
            
        """
        del self.contents[item]
        self.contents = {k: v.remove(item) for k, v in self.contents.items()}
        return
  
    def _disconnect(self, item: base.Edge, *args: Any, **kwargs: Any) -> None:
        """Removes edge from the stored graph.
        
        Args:
            item (Edge): edge to delete from the stored graph.
            
        """
        self.contents[item[0]].remove(item[1])  
        return

    def _merge(self, item: base.Graph, *args: Any, **kwargs: Any) -> None:
        """Combines 'item' with the stored graph.
        
        Args:
            item (Graph): another Graph object to add to the stored graph.
                
        """
        other = base.transform(
            item = item, 
            output = 'adjacency', 
            raise_same_error = False)
        for node, edges in other.items():
            if node in self:
                self[node].update(edges)
            else:
                self[node] = edges
        return
    
    def _subset(
        self, 
        include: Union[Hashable, Sequence[Hashable]] = None,
        exclude: Union[Hashable, Sequence[Hashable]] = None) -> Adjacency:
        """Returns a new graph without a subset of 'contents'.

        Args:
            include (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should be included in the new graph.
            exclude (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should not be included in the new graph.

        Returns:
           Adjacency: with only selected nodes and edges.
            
        """
        if include:
            excludables = [k for k in self.contents if k not in include]
        else:
            excludables = []
        excludables.extend([i for i in self.contents if i in exclude])
        new_graph = copy.deepcopy(self)
        for node in camina.iterify(excludables):
            new_graph.delete(node = node)
        return new_graph  
          
    """ Dunder Methods """
    
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        """Returns whether 'instance' meets criteria to be a subclass.

        Args:
            instance (object): item to test as an instance.

        Returns:
            bool: whether 'instance' meets criteria to be a subclass.
            
        """
        return check.is_adjacency(item = instance)


@dataclasses.dataclass
class Edges(base.Graph, camina.Listing):
    """Base class for edge-list graphs.

    Args:
        contents (MutableSequence[base.Edge]): Listing of edges. Defaults to 
            an empty list.
                                      
    """   
    contents: MutableSequence[base.Edge] = dataclasses.field(
        default_factory = list)

    """ Private Methods """   
    
    def _add(self, item: base.Edge, *args: Any, **kwargs: Any) -> None:
        """Adds edge to the stored graph.
                   
        Args:
            item (base.Edge): edge to add to the stored graph.
            
        """
        self.contents.append(item)
        return
            
    def _connect(self, item: base.Edge, *args: Any, **kwargs: Any) -> None:
        """Adds edge to the stored graph.
        
        Args:
            item (Edge): edge to add to the stored graph.
            
        """
        self.contents.append(item)
        return

    def _delete(self, item: base.Edge, *args: Any, **kwargs: Any) -> None:
        """Removes edge from the stored graph.
        
        Args:
            item (Edge): edge to delete from the stored graph.
            
        """
        self.contents.remove(item) 
        return    
            
    def _disconnect(self, item: base.Edge, *args: Any, **kwargs: Any) -> None:
        """Removes edge from the stored graph.
        
        Args:
            item (Edge): edge to delete from the stored graph.
            
        """
        self.contents.remove(item) 
        return    

    def _merge(self, item: base.Graph, *args: Any, **kwargs: Any) -> None:
        """Combines 'item' with the stored graph.
        
        Args:
            item (Graph): another Graph object to add to the stored graph.
                
        """
        other = base.transform(
            item = item, 
            output = 'edges', 
            raise_same_error = False)
        self.contents.extend(other)
        return
    
    def _subset(
        self, 
        include: Union[Hashable, Sequence[Hashable]] = None,
        exclude: Union[Hashable, Sequence[Hashable]] = None) -> Adjacency:
        """Returns a new graph without a subset of 'contents'.

        Args:
            include (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should be included in the new graph.
            exclude (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should not be included in the new graph.

        Returns:
           Adjacency: with only selected nodes and edges.
            
        """
        raise NotImplementedError 
                              
    """ Dunder Methods """
           
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        """Returns whether 'instance' meets criteria to be a subclass.

        Args:
            instance (object): item to test as an instance.

        Returns:
            bool: whether 'instance' meets criteria to be a subclass.
            
        """
        return check.is_edges(item = instance)

   
@dataclasses.dataclass
class Matrix(base.Graph, camina.Listing):
    """Base class for adjacency-matrix graphs.
    
    Args:
        contents (Sequence[Sequence[int]]): a list of list of integers 
            indicating edges between nodes in the matrix. Defaults to an empty
            list.
        labels (Sequence[Hashable]): names of nodes in the matrix. 
            Defaults to an empty list.
                                      
    """  
    contents: MutableSequence[MutableSequence[int]] = dataclasses.field(
        default_factory = list)
    labels: MutableSequence[Hashable] = dataclasses.field(
        default_factory = list)

    """ Private Methods """   
    
    def _add(self, item: base.Edge, *args: Any, **kwargs: Any) -> None:
        """Adds edge to the stored graph.
                   
        Args:
            item (base.Edge): edge to add to the stored graph.
            
        """
        raise NotImplementedError
            
    def _connect(self, item: base.Edge, *args: Any, **kwargs: Any) -> None:
        """Adds edge to the stored graph.
        
        Args:
            item (Edge): edge to add to the stored graph.
            
        """
        raise NotImplementedError

    def _delete(self, item: base.Edge, *args: Any, **kwargs: Any) -> None:
        """Removes edge from the stored graph.
        
        Args:
            item (Edge): edge to delete from the stored graph.
            
        """
        raise NotImplementedError   
            
    def _disconnect(self, item: base.Edge, *args: Any, **kwargs: Any) -> None:
        """Removes edge from the stored graph.
        
        Args:
            item (Edge): edge to delete from the stored graph.
            
        """
        raise NotImplementedError   

    def _merge(self, item: base.Graph, *args: Any, **kwargs: Any) -> None:
        """Combines 'item' with the stored graph.
        
        Args:
            item (Graph): another Graph object to add to the stored graph.
                
        """
        other = base.transform(
            item = item, 
            output = 'matrix', 
            raise_same_error = False)
        new_matrix = other[0]
        length = len(self.contents)
        for row in enumerate(new_matrix):
            for column in enumerate(row):
                self.contents[row + length][column + length] = (
                    new_matrix[row][column])
        self.labels.extend(other[1])
        return
    
    def _subset(
        self, 
        include: Union[Hashable, Sequence[Hashable]] = None,
        exclude: Union[Hashable, Sequence[Hashable]] = None) -> Adjacency:
        """Returns a new graph without a subset of 'contents'.

        Args:
            include (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should be included in the new graph.
            exclude (Union[Hashable, Sequence[Hashable]]): nodes or edges which 
                should not be included in the new graph.

        Returns:
           Adjacency: with only selected nodes and edges.
            
        """
        raise NotImplementedError 
                             
    """ Dunder Methods """
        
    @classmethod
    def __instancecheck__(cls, instance: object) -> bool:
        """Returns whether 'instance' meets criteria to be a subclass.

        Args:
            instance (object): item to test as an instance.

        Returns:
            bool: whether 'instance' meets criteria to be a subclass.
            
        """
        return check.is_matrix(item = instance)
